# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - MoinMessage library

    @copyright: 2012, 2013 by Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from email import message_from_string
from email.encoders import encode_noop
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.mime.base import MIMEBase
from email.utils import formatdate, parsedate
from subprocess import Popen, PIPE
from tempfile import mkstemp
from urlparse import urlsplit
import httplib
import os

def is_collection(message):
    return message.get("Update-Type") == "collection"

def to_replace(message):
    return message.get("Update-Action") == "replace"

def to_store(message):
    return message.get("Update-Action") == "store"

class Message:

    "An update message."

    def __init__(self, text=None):
        self.date = None
        self.updates = []
        if text:
            self.parse_text(text)

    def init_date(self, message):

        "Obtain the date of the given 'message'."

        if message.has_key("Date"):
            self.date = parsedate(message["Date"])
        else:
            self.date = None

    def parse_text(self, text):

        "Parse the given 'text' as a message."

        self.handle_message(message_from_string(text))

    def handle_message(self, message):

        "Handle the given 'message', recording the separate updates."

        self.init_date(message)

        # The message either consists of a collection of updates.

        if message.is_multipart() and is_collection(message):
            for part in message.get_payload():
                self.updates.append(part)

        # Or the message is a single update.

        else:
            self.updates.append(message)

    def add_updates(self, parts):

        """
        Add the given 'parts' to a message.
        """

        for part in updates:
            self.add_update(part)

    def add_update(self, part):

        """
        Add an update 'part' to a message.
        """

        self.updates.append(part)

    def get_update(self, alternatives):

        """
        Return a suitable multipart object containing the supplied
        'alternatives'.
        """

        part = MIMEMultipart()
        for alternative in alternatives:
            part.attach(alternative)
        return part

    def get_payload(self, timestamped=True):

        """
        Get the multipart payload for the message. If the 'timestamped'
        parameter is omitted or set to a true value, the payload will be given a
        date header set to the current date and time that can be used to assess
        the validity of a message and to determine whether it has already been
        received by a recipient.
        """

        if len(self.updates) == 1:
            message = self.updates[0]
        else:
            message = MIMEMultipart()
            message.add_header("Update-Type", "collection")
            for update in self.updates:
                message.attach(update)

        if timestamped:
            timestamp(message)
            self.init_date(message)

        return message

class Mailbox:

    "A collection of messages within a multipart message."

    def __init__(self, text=None):
        self.messages = []
        if text:
            self.parse_text(text)

    def parse_text(self, text):

        "Parse the given 'text' as a mailbox."

        message = message_from_string(text)

        if message.is_multipart():
            for part in message.get_payload():
                self.messages.append(part)
        else:
            self.messages.append(message)

    def add_message(self, message):

        "Add the given 'message' to the mailbox."

        self.messages.append(message)

    def get_payload(self):

        "Get the multipart payload for the mailbox."

        mailbox = MIMEMultipart()
        for message in self.messages:
            mailbox.attach(message)

        return mailbox

class MoinMessageError(Exception):
    pass

class MoinMessageDecodingError(Exception):
    pass

class MoinMessageMissingPart(MoinMessageDecodingError):
    pass

class MoinMessageBadContent(MoinMessageDecodingError):
    pass

class GPG:

    "A wrapper around the gpg command using a particular configuration."

    def __init__(self, homedir=None):
        self.conf_args = []

        if homedir:
            self.conf_args += ["--homedir", homedir]

        self.errors = None

    def run(self, args, text=None):

        """
        Invoke gpg with the given 'args', supplying the given 'text' to the
        command directly or, if 'text' is omitted, using a file provided as part
        of the 'args' if appropriate.

        Failure to complete the operation will result in a MoinMessageError
        being raised.
        """

        cmd = Popen(["gpg"] + self.conf_args + list(args), stdin=PIPE, stdout=PIPE, stderr=PIPE)

        try:
            # Attempt to write input to the command and to read output from the
            # command.

            try:
                if text:
                    cmd.stdin.write(text)
                cmd.stdin.close()

                text = cmd.stdout.read()

            # I/O errors can indicate the failure of the command.

            except IOError:
                pass

            self.errors = cmd.stderr.read()

            # Test for a zero result.

            if not cmd.wait():
                return text
            else:
                raise MoinMessageError, self.errors

        finally:
            cmd.stdout.close()
            cmd.stderr.close()

    def verifyMessageText(self, signature, content):

        "Using the given 'signature', verify the given message 'content'."

        # Write the detached signature and content to files.

        signature_fd, signature_filename = mkstemp()
        content_fd, content_filename = mkstemp()

        try:
            signature_fp = os.fdopen(signature_fd, "w")
            content_fp = os.fdopen(content_fd, "w")
            try:
                signature_fp.write(signature)
                content_fp.write(content)
            finally:
                signature_fp.close()
                content_fp.close()

            # Verify the message text.

            text = self.run(["--status-fd", "1", "--verify", signature_filename, content_filename])

            # Return the details of the signing key.

            identity = None
            fingerprint = None

            for line in text.split("\n"):
                try:
                    prefix, msgtype, digest, details = line.strip().split(" ", 3)
                except ValueError:
                    continue

                # Return the fingerprint and identity details.

                if msgtype == "GOODSIG":
                    identity = details
                elif msgtype == "VALIDSIG":
                    fingerprint = digest

                if identity and fingerprint:
                    return fingerprint, identity

            return None

        finally:
            os.remove(signature_filename)
            os.remove(content_filename)

    def verifyMessage(self, message):

        """
        Verify the given RFC 3156 'message', returning a tuple of the form
        (fingerprint, identity, content).
        """

        content, signature = getContentAndSignature(message)

        # Verify the message format.

        if signature.get_content_type() != "application/pgp-signature":
            raise MoinMessageBadContent

        # Verify the message.

        fingerprint, identity = self.verifyMessageText(signature.get_payload(), content.as_string())
        return fingerprint, identity, content

    def signMessage(self, message, keyid):

        """
        Return a signed version of 'message' using the given 'keyid'.
        """

        text = message.as_string()
        signature = self.run(["--armor", "-u", keyid, "--detach-sig"], text)

        # Make the container for the message.

        signed_message = MIMEMultipart("signed", protocol="application/pgp-signature")
        signed_message.attach(message)

        signature_part = MIMEBase("application", "pgp-signature")
        signature_part.set_payload(signature)
        signed_message.attach(signature_part)

        return signed_message

    def decryptMessageText(self, message):

        "Return a decrypted version of 'message'."

        return self.run(["--decrypt"], message)

    def decryptMessage(self, message):

        """
        Decrypt the given RFC 3156 'message', returning the message text.
        """

        try:
            declaration, content = message.get_payload()
        except ValueError:
            raise MoinMessageMissingPart

        # Verify the message format.

        if content.get_content_type() != "application/octet-stream":
            raise MoinMessageBadContent

        # Return the decrypted message text.

        return self.decryptMessageText(content.get_payload())

    def encryptMessage(self, message, keyid):

        """
        Return an encrypted version of 'message' using the given 'keyid'.
        """

        text = message.as_string()
        encrypted = self.run(["--armor", "-r", keyid, "--encrypt", "--trust-model", "always"], text)

        # Make the container for the message.

        encrypted_message = MIMEMultipart("encrypted", protocol="application/pgp-encrypted")

        # For encrypted content, add the declaration and content.

        declaration = MIMEBase("application", "pgp-encrypted")
        declaration.set_payload("Version: 1")
        encrypted_message.attach(declaration)

        content = MIMEApplication(encrypted, "octet-stream", encode_noop)
        encrypted_message.attach(content)

        return encrypted_message

# Message decoding functions.

# Detect PGP/GPG-encoded payloads.
# See: http://tools.ietf.org/html/rfc3156

def is_signed(message):
    mimetype = message.get_content_type()
    encoding = message.get_content_charset()

    return mimetype == "multipart/signed" and \
        message.get_param("protocol") == "application/pgp-signature"

def is_encrypted(message):
    mimetype = message.get_content_type()
    encoding = message.get_content_charset()

    return mimetype == "multipart/encrypted" and \
        message.get_param("protocol") == "application/pgp-encrypted"

def getContentAndSignature(message):

    """
    Return the content and signature parts of the given RFC 3156 'message'.

    NOTE: RFC 3156 states that signed messages should employ a detached
    NOTE: signature but then shows "BEGIN PGP MESSAGE" for signatures
    NOTE: instead of "BEGIN PGP SIGNATURE".
    NOTE: The "micalg" parameter is currently not supported.
    """

    try:
        content, signature = message.get_payload()
        return content, signature
    except ValueError:
        raise MoinMessageMissingPart

# Communications functions.

def timestamp(message):

    """
    Timestamp the given 'message' so that its validity can be assessed by the
    recipient.
    """

    datestr = formatdate()

    if not message.has_key("Date"):
        message.add_header("Date", datestr)
    else:
        message["Date"] = datestr

def sendMessage(message, url):

    "Send 'message' to the given 'url."

    scheme, host, port, path = parseURL(url)
    text = message.as_string()

    if scheme == "http":
        cls = httplib.HTTPConnection
    elif scheme == "https":
        cls = httplib.HTTPSConnection
    else:
        raise MoinMessageError, "Communications protocol not supported: %s" % scheme

    req = cls(host, port)
    req.request("PUT", path, text)
    resp = req.getresponse()

    if resp.status >= 400:
        raise MoinMessageError, "Message sending failed: %s" % resp.status

    return resp.read()

def parseURL(url):

    "Return the scheme, host, port and path for the given 'url'."

    scheme, host_port, path, query, fragment = urlsplit(url)
    host_port = host_port.split(":")

    if query:
        path += "?" + query

    if len(host_port) > 1:
        host = host_port[0]
        port = int(host_port[1])
    else:
        host = host_port[0]
        port = 80

    return scheme, host, port, path

# vim: tabstop=4 expandtab shiftwidth=4
